import os
from setuptools import setup, Extension, find_packages

def my_build_ext(pars):
    # import delayed:
    from setuptools.command.build_ext import build_ext as _build_ext
    class build_ext(_build_ext):
        def finalize_options(self):
            # got error `'dict' object has no attribute '__NUMPY_SETUP__'`
            def _set_builtin(name, value):
                if isinstance(__builtins__, dict):
                    __builtins__[name] = value
                else:
                    setattr(__builtins__, name, value)
            _build_ext.finalize_options(self)
            # Prevent numpy from thinking it is still in its setup process:
            _set_builtin('__NUMPY_SETUP__', False)
            import numpy
            self.include_dirs.append(numpy.get_include())
    #object returned:
    return build_ext(pars)

here = os.path.abspath(os.path.dirname(__file__))
# Get the long description from the README file
with open(os.path.join(here, 'README.md')) as f:
    long_description = f.read()

if os.name == 'nt':
    extra_compile_args = ["-Ox"]
else:
    extra_compile_args = ['-std=c++0x', '-pthread', '-O3']

array_wrappers_ext = Extension('sparse_dot_topn.array_wrappers',
                         sources=[
                                    './sparse_dot_topn/array_wrappers.pyx',
                                ],
                         extra_compile_args=extra_compile_args,
                         language='c++')

original_ext = Extension('sparse_dot_topn.sparse_dot_topn',
                         sources=[
                                    './sparse_dot_topn/sparse_dot_topn.pyx',
                                    './sparse_dot_topn/sparse_dot_topn_source.cpp'
                                ],
                         extra_compile_args=extra_compile_args,
                         language='c++')

threaded_ext = Extension('sparse_dot_topn.sparse_dot_topn_threaded',
                         sources=[
                             './sparse_dot_topn/sparse_dot_topn_threaded.pyx',
                             './sparse_dot_topn/sparse_dot_topn_source.cpp',
                             './sparse_dot_topn/sparse_dot_topn_parallel.cpp'],
                         extra_compile_args=extra_compile_args,
                         language='c++')

VERSION = '0.0.1'
DESCRIPTION = 'compute the cosinus similarity comparison between 2 columns of a sparse matrix very fast and selecting only the top-n best scores with a threshold'

# Setting up
setup(
    name="cosim_comparison_sparse_matrix",
    version="0.0.1",
    author="camillebrl (Camille Barboule)",
    author_email="camille.barboule@gmail.com",
    description="compute the cosinus similarity comparison between 2 columns of a sparse matrix very fast and selecting only the top-n best scores with a threshold",
    long_description_content_type="text/markdown",
    setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        'setuptools>=42',
        'cython>=0.29.15',
        'numpy>=1.16.6', # select this version for Py2/3 compatible
        'scipy>=1.2.3'   # select this version for Py2/3 compatible
    ],
    install_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        'setuptools>=42',
        'cython>=0.29.15',
        'numpy>=1.16.6', # select this version for Py2/3 compatible
        'scipy>=1.2.3'   # select this version for Py2/3 compatible
    ],
    zip_safe=False,
    packages=find_packages(),
    cmdclass={'build_ext': my_build_ext},
    ext_modules=[array_wrappers_ext, original_ext, threaded_ext],
    package_data = {
        'cosim_comparison_spase_matrix': ['./cosim_comparison_spase_matrix/*.pxd']
    },
    include_package_data=True,
)


import sys
import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse import isspmatrix_csr


import cosim_comparison_sparse_matrix as ct
import sparse_dot_topn_threaded as ct_thread


def awesome_cossim_topn(
        A, B, ntop, lower_bound=0, use_threads=False, n_jobs=1, return_best_ntop=False, test_nnz_max=-1):
    """
    This function will return a matrix C in CSR format, where
    C = [sorted top n results > lower_bound for each row of A * B].
    If return_best_ntop=True then best_ntop
    (the true maximum number of elements > lower_bound per row of A * B)
    will also be returned in a tuple together with C as (C, best_ntop).

    Input:
        A and B: two CSR matrices
        ntop: top n results
        lower_bound: a threshold that the element of A*B must be greater than
        use_threads: use multi-thread or not
        n_jobs: number of thread, must be >= 1
        return_best_ntop: (default: False) if True, will return best_ntop together 
                          with C as a tuple: (C, best_ntop)

    Output:
        C: result matrix (returned alone, if return_best_ntop=False)
        best_ntop: The true maximum number of elements > lower_bound per row of 
                   A * B returned together with C as a tuple: (C, best_ntop). It is 
                   returned only if return_best_ntop=True.

    N.B. if A and B are not in CSR format, they will be converted to CSR
    """
    def try_malloc(sz: int, idx_dtype, data_dtype) -> bool:
        try:
            ind_arr = np.empty(sz, dtype=idx_dtype)
            dat_arr = np.empty(sz, dtype=data_dtype)
            del ind_arr, dat_arr
            return True
        except MemoryError:
            return False
        
    if not isspmatrix_csr(A):
        A = A.tocsr()
    if not isspmatrix_csr(B):
        B = B.tocsr()

    dtype = A.dtype
    assert B.dtype == dtype
    lower_bound = dtype.type(lower_bound)  # Casting this scalar to the same type

    M, K1 = A.shape
    K2, N = B.shape

    if K1 != K2:
        err_str = 'A matrix multiplication will be operated. A.shape[1] must be equal to B.shape[0]!'
        raise ValueError(err_str)

    idx_dtype = np.int32

    nnz_max = M*ntop

    # basic check. if A or B are all zeros matrix, return all zero matrix directly
    if len(A.indices) == 0 or len(B.indices) == 0:
        indptr = np.zeros(M + 1, dtype=idx_dtype)
        indices = np.zeros(nnz_max, dtype=idx_dtype)
        data = np.zeros(nnz_max, dtype=A.dtype)
        output = csr_matrix((data, indices, indptr), shape=(M, N))
        if return_best_ntop:
            return output, 0
        else:
            return output

    indptr = np.empty(M + 1, dtype=idx_dtype)

    # reduce nnz_max if too large to fit in available memory:
    nnz_max = 16*nnz_max
    while (not try_malloc(nnz_max, idx_dtype, A.dtype)):
        nnz_max = nnz_max//2

    # take a chance on high matrix-sparsity and reduce further:
    nnz_max = max(M, nnz_max//16)
    
    # this line is only for testing purposes, designed to enable the user to 
    # force C/C++ to reallocate memory during the matrix multiplication
    nnz_max = test_nnz_max if test_nnz_max > 0 else nnz_max
    
    # filled matrices from here on
    indices = np.empty(nnz_max, dtype=idx_dtype)
    data = np.empty(nnz_max, dtype=A.dtype)

    best_ntop_arr = np.full(1, 0, dtype=idx_dtype)

    if not use_threads:

        alt_indices, alt_data = ct.sparse_dot_topn_extd(
            M, N, np.asarray(A.indptr, dtype=idx_dtype),
            np.asarray(A.indices, dtype=idx_dtype),
            A.data,
            np.asarray(B.indptr, dtype=idx_dtype),
            np.asarray(B.indices, dtype=idx_dtype),
            B.data,
            ntop,
            lower_bound,
            indptr, indices, data, best_ntop_arr
        )

    else:
        if n_jobs < 1:
            err_str = 'Whenever you select the multi-thread mode, n_job must be greater than or equal to 1!'
            raise ValueError(err_str)

        alt_indices, alt_data = ct_thread.sparse_dot_topn_extd_threaded(
            M, N, np.asarray(A.indptr, dtype=idx_dtype),
            np.asarray(A.indices, dtype=idx_dtype),
            A.data,
            np.asarray(B.indptr, dtype=idx_dtype),
            np.asarray(B.indices, dtype=idx_dtype),
            B.data,
            ntop,
            lower_bound,
            indptr, indices, data, best_ntop_arr, n_jobs
        )

    if alt_indices is not None:
        indices = alt_indices
        data = alt_data

    # prepare and return the output:
    output = csr_matrix((data, indices, indptr), shape=(M, N))
    if return_best_ntop:
        return output, best_ntop_arr[0]
    else:
        return output
